# -*- coding: utf-8 -*-
"""hgvs.normalizer
"""

import copy
import logging

from bioutils.sequences import reverse_complement

import hgvs
import hgvs.validator
import hgvs.variantmapper
from hgvs.exceptions import (
    HGVSDataNotAvailableError,
    HGVSInvalidVariantError,
    HGVSUnsupportedOperationError,
)
from hgvs.utils.norm import normalize_alleles

_logger = logging.getLogger(__name__)


class Normalizer:
    """Perform variant normalization"""

    def __init__(
        self,
        hdp,
        cross_boundaries=hgvs.global_config.normalizer.cross_boundaries,
        shuffle_direction=hgvs.global_config.normalizer.shuffle_direction,
        alt_aln_method=hgvs.global_config.mapping.alt_aln_method,
        validate=hgvs.global_config.normalizer.validate,
        variantmapper=None,
    ):
        """Initialize and configure the normalizer

        :param hdp: HGVS Data Provider Interface-compliant instance
            (see :class:`hgvs.dataproviders.interface.Interface`)
        :param cross_boundaries: whether allow the shuffling to cross the exon-intron boundary
        :param shuffle_direction: shuffling direction
        :param alt_aln_method: sequence alignment method (e.g., splign, blat)
        :param validate: whether validating the input variant before normalizing

        """
        assert (
            shuffle_direction == 3 or shuffle_direction == 5
        ), "The shuffling direction should be 3 (3' most) or 5 (5' most)."
        self.hdp = hdp
        self.shuffle_direction = shuffle_direction
        self.cross_boundaries = cross_boundaries
        self.alt_aln_method = alt_aln_method
        self.validator = None
        if validate:
            self.validator = hgvs.validator.IntrinsicValidator(strict=False)
        self.vm = variantmapper or hgvs.variantmapper.VariantMapper(self.hdp)

    def normalize(self, var):
        """Perform sequence variants normalization for single variant"""
        assert isinstance(
            var, hgvs.sequencevariant.SequenceVariant
        ), "variant must be a parsed HGVS sequence variant object"

        # keep a shallow reference to the original variant, to be returned
        # as-is under certain circumstances
        orig_var = var

        if self.validator:
            self.validator.validate(var)

        init_met = False
        if var.posedit is not None and isinstance(var.posedit, hgvs.edit.AARefAlt):
            init_met = var.posedit.init_met

        if var.posedit is None or var.posedit.uncertain or init_met or var.posedit.pos is None:
            return var

        type = var.type

        if type == "p":
            raise HGVSUnsupportedOperationError(
                "Unsupported normalization of protein level variants: {0}".format(var)
            )
        if var.posedit.edit.type == "con":
            raise HGVSUnsupportedOperationError(
                "Unsupported normalization of conversion variants: {0}", format(var)
            )

        var.fill_ref(self.hdp)

        if var.posedit.edit.type == "identity":
            var_norm = copy.deepcopy(var)
            return var_norm

        # For c. variants normalization, first convert to n. variant
        # and perform normalization at the n. level, then convert the
        # normalized n. variant back to c. variant.
        if type == "c":
            var = self.vm.c_to_n(var)

        if var.type in "nr":
            if var.posedit.pos.start.offset != 0 or var.posedit.pos.end.offset != 0:
                raise HGVSUnsupportedOperationError(
                    "Normalization of intronic variants is not supported"
                )

        def is_valid_pos(ac, pos):
            # tests whether the sequence position actually exists
            # This is *way* janky.
            # TODO: push functionality to hdp which can implement differently
            # based on capabilities of sequence backend
            try:
                s = self.hdp.get_seq(ac, pos - 1, pos)  # 0-based!
                return s != ""
            except HGVSDataNotAvailableError as e:
                # Bad Request indicates that we got to NCBI, but the request
                # was invalid.
                return "Bad Request" not in str(e)

        if var.posedit.pos.start.base < 0 or not is_valid_pos(var.ac, var.posedit.pos.end.base):
            if hgvs.global_config.mapping.strict_bounds:
                raise HGVSInvalidVariantError(f"{var}: coordinates are out-of-bounds")
            _logger.warning(f"{var}: coordinates are out-of-bounds; returning as-is")
            return orig_var

        # restrict var types to those that use sequence start (i.e., not c.)
        assert var.type in "gmnr", "Internal Error: variant must be of type g, m, n, r"

        bound_s, bound_e = self._get_boundary(var)
        boundary = (bound_s, bound_e)
        start, end, (ref, alt) = self._normalize_alleles(var, boundary)

        ref_len = len(ref)
        alt_len = len(alt)

        # Generate normalized variant
        if alt_len == ref_len:
            ref_start = start
            ref_end = end - 1
            # inversion
            if ref_len > 1 and ref == reverse_complement(alt):
                edit = hgvs.edit.Inv(ref=ref)
            # ident
            elif ref_len == 0 and alt_len == 0:
                ref_start = ref_end
                edit = hgvs.edit.NARefAlt(ref=ref, alt=alt)
            # substitution or delins
            else:
                edit = hgvs.edit.NARefAlt(ref=ref, alt=alt)
        if alt_len < ref_len:
            # del or delins
            ref_start = start
            ref_end = end - 1
            edit = hgvs.edit.NARefAlt(ref=ref, alt=None if alt_len == 0 else alt)
        elif alt_len > ref_len:
            # ins or dup
            if ref_len == 0:
                if self.shuffle_direction == 3:
                    adj_seq = self._fetch_bounded_seq(
                        var, start - alt_len - 1, end - 1, 0, boundary
                    )
                else:
                    adj_seq = self._fetch_bounded_seq(
                        var, start - 1, start + alt_len - 1, 0, boundary
                    )
                # ins
                if alt != adj_seq:
                    ref_start = start - 1
                    ref_end = end
                    edit = hgvs.edit.NARefAlt(ref=None, alt=alt)
                # dup
                else:
                    if self.shuffle_direction == 3:
                        ref_start = start - alt_len
                        ref_end = end - 1
                        edit = hgvs.edit.Dup(ref=alt)
                    else:
                        ref_start = start
                        ref_end = start + alt_len - 1
                        edit = hgvs.edit.Dup(ref=alt)
            # delins
            else:
                ref_start = start
                ref_end = end - 1
                edit = hgvs.edit.NARefAlt(ref=ref, alt=alt)

        # ensure the start is not 0
        if ref_start == 0:
            ref = self._fetch_bounded_seq(var, 0, 1, 0, boundary)
            alt = alt + ref
            edit = hgvs.edit.NARefAlt(ref=ref, alt=alt)
            ref_start = 1
            ref_end = 1

        # ensure the end is not outside of reference sequence
        tgt_len = self._get_tgt_length(var)
        if ref_end == tgt_len + 1:
            ref = self._fetch_bounded_seq(var, tgt_len - 1, tgt_len, 0, boundary)
            alt = ref + alt
            edit = hgvs.edit.NARefAlt(ref=ref, alt=alt)
            ref_start = tgt_len
            ref_end = tgt_len

        var_norm = copy.deepcopy(var)
        var_norm.posedit.edit = edit
        var_norm.posedit.pos.start.base = ref_start
        var_norm.posedit.pos.end.base = ref_end

        if type == "c":
            var_norm = self.vm.n_to_c(var_norm)

        return var_norm

    def _get_boundary(self, var):
        """Get the position of exon-intron boundary for current variant"""
        if var.type == "r" or var.type == "n":
            if self.cross_boundaries:
                return 0, float("inf")
            else:
                # Get genomic sequence access number for this transcript
                map_info = self.hdp.get_tx_mapping_options(var.ac)
                if not map_info:
                    raise HGVSDataNotAvailableError(
                        "No mapping info available for {ac}".format(ac=var.ac)
                    )
                map_info = [
                    item for item in map_info if item["alt_aln_method"] == self.alt_aln_method
                ]
                alt_ac = map_info[0]["alt_ac"]

                # Get tx info
                tx_info = self.hdp.get_tx_info(var.ac, alt_ac, self.alt_aln_method)
                cds_start = tx_info["cds_start_i"]
                cds_end = tx_info["cds_end_i"]

                # Get exon info
                exon_info = self.hdp.get_tx_exons(var.ac, alt_ac, self.alt_aln_method)
                exon_starts = [exon["tx_start_i"] for exon in exon_info]
                exon_ends = [exon["tx_end_i"] for exon in exon_info]
                exon_starts.sort()
                exon_ends.sort()
                exon_starts.append(exon_ends[-1])
                exon_ends.append(float("inf"))

                # Find the end pos of the exon where the var locates
                left = 0
                right = float("inf")

                # TODO: #242: implement methods to find tx regions
                for i, _ in enumerate(exon_starts):
                    if (
                        var.posedit.pos.start.base - 1 >= exon_starts[i]
                        and var.posedit.pos.start.base - 1 < exon_ends[i]
                    ):
                        break

                for j, _ in enumerate(exon_starts):
                    if (
                        var.posedit.pos.end.base - 1 >= exon_starts[j]
                        and var.posedit.pos.end.base - 1 < exon_ends[j]
                    ):
                        break

                if i != j:
                    raise HGVSUnsupportedOperationError(
                        "Unsupported normalization of variants spanning the exon-intron boundary ({var})".format(
                            var=var
                        )
                    )

                left = exon_starts[i]
                right = exon_ends[i]

                if cds_start is None:
                    pass
                elif var.posedit.pos.end.base - 1 < cds_start:
                    right = min(right, cds_start)
                elif var.posedit.pos.start.base - 1 >= cds_start:
                    left = max(left, cds_start)
                else:
                    raise HGVSUnsupportedOperationError(
                        "Unsupported normalization of variants spanning the UTR-exon boundary ({var})".format(
                            var=var
                        )
                    )

                if cds_end is None:
                    pass
                elif var.posedit.pos.start.base - 1 >= cds_end:
                    left = max(left, cds_end)
                elif var.posedit.pos.end.base - 1 < cds_end:
                    right = min(right, cds_end)
                else:
                    raise HGVSUnsupportedOperationError(
                        "Unsupported normalization of variants spanning the exon-UTR boundary ({var})".format(
                            var=var
                        )
                    )

                return left, right
        else:
            # For variant type of g and m etc.
            return 0, float("inf")

    def _get_tgt_length(self, var):
        """Get the total length of the whole reference sequence"""
        if var.type == "g" or var.type == "m":
            return float("inf")
        else:
            # Get genomic sequence access number for this transcript
            identity_info = self.hdp.get_tx_identity_info(var.ac)
            if not identity_info:
                raise HGVSDataNotAvailableError(
                    "No identity info available for {ac}".format(ac=var.ac)
                )
            tgt_len = sum(identity_info["lengths"])
            return tgt_len

    def _fetch_bounded_seq(self, var, start, end, window_size, boundary):
        """Fetch reference sequence from hgvs data provider.

        The start position is 0 and the interval is half open
        """
        var_len = end - start - window_size

        start = start if start >= boundary[0] else boundary[0]
        end = end if end <= boundary[1] else boundary[1]
        if start >= end:
            return ""

        seq = self.hdp.get_seq(var.ac, start, end)

        if len(seq) < end - start and len(seq) < var_len:
            raise HGVSInvalidVariantError(
                "Variant span is outside sequence bounds ({var})".format(var=var)
            )

        return seq

    def _get_ref_alt(self, var, boundary):
        """Get reference allele and alternative allele of the variant"""

        # Get reference allele
        if var.posedit.edit.type == "ins" or var.posedit.edit.type == "dup":
            ref = ""
        else:
            # For NARefAlt and Inv
            if var.posedit.edit.ref_s is None or var.posedit.edit.ref == "":
                ref = self._fetch_bounded_seq(
                    var, var.posedit.pos.start.base - 1, var.posedit.pos.end.base, 0, boundary
                )
            else:
                ref = var.posedit.edit.ref

        # Get alternative allele
        if (
            var.posedit.edit.type == "sub"
            or var.posedit.edit.type == "delins"
            or var.posedit.edit.type == "ins"
        ):
            alt = var.posedit.edit.alt
        elif var.posedit.edit.type == "del":
            alt = ""
        elif var.posedit.edit.type == "dup":
            alt = var.posedit.edit.ref or self._fetch_bounded_seq(
                var, var.posedit.pos.start.base - 1, var.posedit.pos.end.base, 0, boundary
            )
        elif var.posedit.edit.type == "inv":
            alt = reverse_complement(ref)
        elif var.posedit.edit.type == "identity":
            alt = ref

        return ref, alt

    def _normalize_alleles(self, var, boundary):
        """Normalize the variant until it could not be shuffled"""

        ref, alt = self._get_ref_alt(var, boundary)
        win_size = hgvs.global_config.normalizer.window_size

        if self.shuffle_direction == 3:
            if var.posedit.edit.type == "ins":
                base = var.posedit.pos.start.base
                start = 1
                stop = 1
            elif var.posedit.edit.type == "dup":
                base = var.posedit.pos.end.base
                start = 1
                stop = 1
            else:
                base = var.posedit.pos.start.base
                start = 0
                stop = var.posedit.pos.end.base - base + 1

            while True:
                ref_seq = self._fetch_bounded_seq(
                    var, base - 1, base + stop - 1 + win_size, win_size, boundary
                )
                if ref_seq == "":
                    break
                orig_start, orig_stop = start, stop
                start, stop, (ref, alt) = normalize_alleles(
                    ref_seq, start, stop, (ref, alt), len(ref_seq), win_size, False
                )
                if stop < len(ref_seq) or start == orig_start:
                    break
                # if stop at the end of the window, try to extend the shuffling to the right
                base += start - orig_start
                stop -= start - orig_start
                start = orig_start

        elif self.shuffle_direction == 5:
            if var.posedit.edit.type == "ins":
                base = max(var.posedit.pos.start.base - win_size, 1)
                start = var.posedit.pos.end.base - base
                stop = var.posedit.pos.end.base - base
            elif var.posedit.edit.type == "dup":
                base = max(var.posedit.pos.start.base - win_size, 1)
                start = var.posedit.pos.end.base - base + 1
                stop = var.posedit.pos.end.base - base + 1
            else:
                base = max(var.posedit.pos.start.base - win_size, 1)
                start = var.posedit.pos.start.base - base
                stop = var.posedit.pos.end.base - base + 1

            while True:
                if base < boundary[0] + 1:
                    start -= boundary[0] + 1 - base
                    stop -= boundary[0] + 1 - base
                    base = boundary[0] + 1
                ref_seq = self._fetch_bounded_seq(var, base - 1, base + stop - 1, start, boundary)
                if ref_seq == "":
                    break
                orig_start, orig_stop = start, stop
                start, stop, (ref, alt) = normalize_alleles(
                    ref_seq, start, stop, (ref, alt), 0, win_size, True
                )
                if start > 0 or stop == orig_stop:
                    break
                # if stop at the end of the window, try to extend the shuffling to the left
                base -= orig_stop - stop
                start += orig_stop - stop
                stop = orig_stop

        return base + start, base + stop, (ref, alt)

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Represent partial HGVS tags that refer to a position without alleles

"""

import attr


@attr.s(slots=True, repr=False)
class HGVSPosition:
    """
    HGVSPosition -- Represent partial HGVS tags that refer to a position without alleles

    :param str ac: sequence accession
    :param str type: type of sequence and coordinate
    :param str pos: sequence position
    :param str gene: gene symbol (may be None)

    """

    ac = attr.ib()
    type = attr.ib()
    pos = attr.ib()
    gene = attr.ib(default=None)

    def __str__(self):
        g = "" if not self.gene else "(" + self.gene + ")"
        return "{self.ac}{g}:{self.type}.{self.pos}".format(self=self, g=g)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join((a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__),
        )


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
import csv
import gzip
import io
import os
import unittest

import pytest

import hgvs
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.sequencevariant
import hgvs.variantmapper
from support import CACHE
from support.crosschecker import CrossChecker, LineIterator

data_fn = os.path.join(os.path.dirname(__file__), "data", "clinvar.gz")


class Test_Clinvar(unittest.TestCase, CrossChecker):
    @classmethod
    def setUpClass(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp)
        self.hp = hgvs.parser.Parser()

    @pytest.mark.extra
    def test_clinvar(self, fn=data_fn, mod=None):
        """Test genome-transcript projections for 7498 clinvar variants in 4676 against genes
        for both GRCh37 and GRCh38 (when both are available).

        Approximate timings in various configurations:
        uta         sequences        time
        local       remote          1 min
        local       remote         40 min
        remote      local          10 min
        remote      remote         50 min

        In other words, you really want to run this with local sequences and UTA.
        """
        fh = LineIterator(
            fh=(
                io.TextIOWrapper(gzip.open(fn), encoding="utf-8")
                if fn.endswith(".gz")
                else io.open(fn)
            ),
            skip_comments=True,
        )
        for rec in csv.DictReader(fh, delimiter=str("\t")):
            if mod and fh.lines_read % mod != 0:
                continue
            print(rec["gene"])
            variants = (self.hp.parse_hgvs_variant(vs) for vs in rec["hgvs_variants"].split())
            variants = [v for v in variants if v.type in "cg" and not v.ac.startswith("NG")]
            try:
                msg = self.crosscheck_variant_group(list(variants))
            except Exception as e:
                e.args = (e.args[0] + "\n  at line {fh.lines_read}: {fh.last_line}".format(fh=fh),)
                raise e
            self.assertIsNone(
                msg, lambda: msg + "\n   on line {fh.lines_read}: {fh.last_line}".format(fh=fh)
            )


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import os
import unittest

import pytest
from support import CACHE

import hgvs.assemblymapper
import hgvs.dataproviders.uta


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    """
    HGVS-730 fixes an issue where AARefAlt.format() sometimes returned the three letter amino acid
    when the single letter format was requested.
    """

    @classmethod
    def setUpClass(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign"
        )
        self.hp = hgvs.parser.Parser()

    def test_730_format_startloss_as_configured(self):
        """
        Parse a start loss and make sure the one letter and three letter protein genotypes come back as expected.
        """
        var_g = self.hp.parse_hgvs_variant("NC_000016.9:g.89985662_89985667del")
        var_c = self.am37.g_to_c(var_g, str("NM_002386.3"))
        var_p = self.am37.c_to_p(var_c)
        var_p_one_letter = var_p.format(conf={"p_3_letter": False})
        var_p_three_letter = var_p.format(conf={"p_3_letter": True})

        self.assertEqual(var_p_one_letter, "NP_002377.4:p.M1?")
        self.assertEqual(var_p_three_letter, "NP_002377.4:p.Met1?")



# -*- coding: utf-8 -*-
"""
Mapping intervals between pairs of congruent segments

The IntervalMapper class is at the heart of mapping between aligned sequences.  An instance
of :class:`uta.tools.intervalmapper.IntervalMapper` is constructed with an ordered list of
:class:`uta.tools.intervalmapper.IntervalPair` instances, each of which consists of two
:class:`uta.tools.intervalmapper.Interval` instances.  The IntervalMapper class is unaware
of strand/orientation; that issue is handled by the
:class:`uta.tools.transcriptmapper.TranscriptMapper` class.

NOTE: Mapping at the boundaries around indels requires a choice.  If seq B
has an insertion relative to seq A, then mapping coordinate at the
boundaries can either be minimal or maximal for both the start and
end. Consider this alignment::

        0         15   20         35         50
        |==========|====|==========|==========|
        |          | __/        __/|          |
        |          |/          /   |          |
        |==========|==========|====|==========|
        0         15         30   35         50
            15M   5D   15M      5I      15M

  segment 1: [ 0,15] ~ [ 0,15]
  segment 2: [15,20] ~ [15,15]
  segment 3: [20,35] ~ [15,30]
  segment 4: [35,35] ~ [30,35]
  segment 5: [35,50] ~ [35,50]

and these intervals around reference position 35::

  interval 1: 34,36   -> 29,36 (no ambiguity)
  interval 2: 35,35   -> 30,35 (reasonable)
  interval 3: 34,35   -> 29,30 (minimal) or 29,35 (maximal)
  interval 4: 35,36   -> 35,36 (minimal) or 30,36 (maximal)

So, for interval 3, end_i=35 matches segment 3 and segment 4.  Analagously
for interval 4, start_i=35 matches segment 4 and segment 5.

Currently, this code matches an interval <start_i,end_i> using the maximal
start_i and minimal end_i.
"""

import logging
import re

from hgvs.exceptions import HGVSInvalidIntervalError

_logger = logging.getLogger(__name__)
_logger.warning("This module is deprecated and will be removed in a future release")


# N.B. This Interval is internal to intervalmapper.py. It is *NOT* the
# same as the Interval defined in location.py.
class Interval:
    """Represents a segment of a sequence in interbase
    coordinates (0-based, right-open).
    """

    __slots__ = ("start_i", "end_i")

    def __init__(self, start_i, end_i):
        if not (start_i <= end_i):
            raise HGVSInvalidIntervalError("start_i must be less than or equal to end_i")
        self.start_i = start_i
        self.end_i = end_i

    @property
    def len(self):
        return self.end_i - self.start_i

    def __repr__(self):
        return "{self.__class__.__name__}(start_i={self.start_i},end_i={self.end_i})".format(
            self=self
        )


class IntervalPair:
    """Represents a match, insertion, or deletion segment of an
    alignment. If a match, the lengths must be equal; if an insertion or
    deletion, the length of the ref or tgt must be zero respectively."""

    __slots__ = ("ref", "tgt")

    def __init__(self, ref, tgt):
        if not (
            (ref.len == tgt.len)
            or (ref.len == 0 and tgt.len != 0)
            or (ref.len != 0 and tgt.len == 0)
        ):
            raise HGVSInvalidIntervalError(
                "IntervalPair doesn't represent a match, insertion, or deletion"
            )
        self.ref = ref
        self.tgt = tgt

    def __repr__(self):
        return "{self.__class__.__name__}(ref={self.ref},tgt={self.tgt})".format(self=self)


class IntervalMapper:
    """Provides mapping between sequence coordinates according to an
    ordered set of IntervalPairs."""

    __slots__ = ("interval_pairs", "ref_intervals", "tgt_intervals", "ref_len", "tgt_len")

    def __init__(self, interval_pairs):
        """
        :param interval_pairs: an ordered list of IntervalPair instances
        :type interval_pairs: list (of IntervalPair instances).
        :returns: an IntervalMapper instance
        """

        def _validate_intervals(ivs):
            for i in range(1, len(ivs)):
                # check for adjacency/non-overlap
                # This constraint, combined with the start_i <= end_i constraint
                # of Intervals, guarantees that intervals are ordered
                assert ivs[i - 1].end_i == ivs[i].start_i, "intervals must be adjacent"

        self.interval_pairs = interval_pairs
        self.ref_intervals = [ip.ref for ip in self.interval_pairs]
        self.tgt_intervals = [ip.tgt for ip in self.interval_pairs]
        _validate_intervals(self.ref_intervals)
        _validate_intervals(self.tgt_intervals)
        self.ref_len = sum([iv.len for iv in self.ref_intervals])
        self.tgt_len = sum([iv.len for iv in self.tgt_intervals])

    @staticmethod
    def from_cigar(cigar):
        """
        :param cigar: a Compact Idiosyncratic Gapped Alignment Report string
        :type cigar: str.
        :returns: an IntervalMapper instance from the CIGAR string
        """
        return IntervalMapper(cigar_to_intervalpairs(cigar))

    def map_ref_to_tgt(self, start_i, end_i, max_extent=False):
        return self._map(self.ref_intervals, self.tgt_intervals, start_i, end_i, max_extent)

    def map_tgt_to_ref(self, start_i, end_i, max_extent=False):
        return self._map(self.tgt_intervals, self.ref_intervals, start_i, end_i, max_extent)

    @staticmethod
    def _map(from_ivs, to_ivs, from_start_i, from_end_i, max_extent):
        def iv_map(from_ivs, to_ivs, from_start_i, from_end_i, max_extent):
            """returns the <start,end> intervals indexes in which from_start_i and from_end_i occur"""
            # first look for 0-width interval that matches
            seil = [
                i
                for i, iv in enumerate(from_ivs)
                if iv.start_i == from_start_i and iv.end_i == from_end_i
            ]
            if len(seil) > 0:
                si = ei = seil[0]
            else:
                sil = [i for i, iv in enumerate(from_ivs) if iv.start_i <= from_start_i <= iv.end_i]
                eil = [i for i, iv in enumerate(from_ivs) if iv.start_i <= from_end_i <= iv.end_i]
                if len(sil) == 0 or len(eil) == 0:
                    raise HGVSInvalidIntervalError(
                        "start or end or both are beyond the bounds of transcript record"
                    )
                si, ei = (sil[0], eil[-1]) if max_extent else (sil[-1], eil[0])
            return si, ei

        def clip_to_iv(iv, pos):
            return max(iv.start_i, min(iv.end_i, pos))

        assert from_start_i <= from_end_i, "expected from_start_i <= from_end_i"
        try:
            si, ei = iv_map(from_ivs, to_ivs, from_start_i, from_end_i, max_extent)
        except ValueError:
            raise HGVSInvalidIntervalError("start_i,end_i interval out of bounds")
        to_start_i = clip_to_iv(
            to_ivs[si], to_ivs[si].start_i + (from_start_i - from_ivs[si].start_i)
        )
        to_end_i = clip_to_iv(to_ivs[ei], to_ivs[ei].end_i - (from_ivs[ei].end_i - from_end_i))
        return to_start_i, to_end_i


class CIGARElement:
    """represents elements of a CIGAR string and provides methods for
    determining the number of ref and tgt bases consumed by the
    operation"""

    __slots__ = ("len", "op")

    def __init__(self, len, op):
        self.len = len
        self.op = op

    @property
    def ref_len(self):
        """returns number of nt/aa consumed in reference sequence for this edit"""
        return self.len if self.op in "=INX" else 0

    @property
    def tgt_len(self):
        """returns number of nt/aa consumed in target sequence for this edit"""
        return self.len if self.op in "=DX" else 0


def cigar_to_intervalpairs(cigar):
    """For a given CIGAR string, return a list of (Interval,Interval)
    pairs.  The length of the returned list will be equal to the
    number of CIGAR operations
    """

    cigar_elem_re = re.compile(r"(?P<len>\d+)(?P<op>[=DIMNX])")
    ces = [
        CIGARElement(op=md["op"], len=int(md["len"]))
        for md in [m.groupdict() for m in cigar_elem_re.finditer(cigar)]
    ]
    ips = [None] * len(ces)
    ref_pos = tgt_pos = 0
    for i, ce in enumerate(ces):
        ips[i] = IntervalPair(
            Interval(ref_pos, ref_pos + ce.ref_len), Interval(tgt_pos, tgt_pos + ce.tgt_len)
        )
        ref_pos += ce.ref_len
        tgt_pos += ce.tgt_len
    return ips


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
import unittest

import pytest

import hgvs.edit
import hgvs.location
import hgvs.posedit


@pytest.mark.quick
@pytest.mark.models
class Test_PosEdit(unittest.TestCase):
    def test_PosEdit(self):
        pos = hgvs.location.Interval(
            hgvs.location.BaseOffsetPosition(base=12, offset=+34),
            hgvs.location.BaseOffsetPosition(base=56, offset=-78),
        )
        edit = hgvs.edit.NARefAlt("AA", None)
        pe = hgvs.posedit.PosEdit(pos=pos, edit=edit)
        self.assertEqual(pe.format(conf={"max_ref_length": None}), "12+34_56-78delAA")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ChromSeqRendered(BasicRenderer):
    """
    ChromSeqRendered is a class that extends BasicRenderer to provide
    functionality for rendering chromosome sequences with color coding
    for different nucleotides.
    Methods
    -------
    legend() -> str
        Returns a string representing the legend for the sequence
        orientation based on the configuration.
    display(data: VariantData) -> str
        Colors the reference sequences with adenine (A, green), thymine (T, red),
        cytosine (C, yellow), and guanine (G, blue) based on the provided
        VariantData.
    """

    def legend(self) -> str:
        if self.orientation < 0 and self.config.reverse_display:
            arrow = "<-"
        else:
            arrow = "->"

        return f"seq    {arrow} : "

    def display(self, data: VariantData) -> str:
        """colors the ref sequences with adenine (A, green), thymine (T, red), cytosine (C, yellow), and guanine (G, blue)"""
        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        var_seq = ""
        for p in data.position_details:
            c = p.ref

            if not c:
                var_seq += "."
                continue

            if self.config.use_color:
                var_seq += COLOR_MAP.get(c, COLOR_MAP["N"])

            var_seq += c

            if self.config.use_color:
                var_seq += ENDC

        return var_seq



# -*- coding: utf-8 -*-
import hashlib
import os
import pprint
import re
import unittest

import pytest

import hgvs.parser
from hgvs.exceptions import HGVSParseError


def test_parser_variants_with_gene_names(parser):
    assert parser.parse("NM_01234.5(BOGUS):c.22+1A>T")

    # HGNC approved symbols include non-alphanumeric:
    # dashes     - ADAMTSL4-AS1 or TRL-CAA5-1
    assert parser.parse("NM_01234.5(BOGUS-EXCELLENT):c.22+1A>T")
    assert parser.parse("NM_01234.5(BOGUS-MOST-EXCELLENT):c.22+1A>T")

    # underscore - GTF2H2C_2, APOBEC3A_B, C4B_2
    assert parser.parse("NM_01234.5(BOGUS_EXCELLENT):c.22+1A>T")

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("NM_01234.5(1BOGUS):c.22+1A>T")  # Starts with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("NM_01234.5(-BOGUS):c.22+1A>T")  # Starts with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("NM_01234.5(BOGUS-):c.22+1A>T")  # Ends with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("NM_01234.5(BOGUS/EXCELLENT):c.22+1A>T")  # contains invalid character


def test_parser_variants_with_no_transcript_gene_names(parser):
    """Test it also works with no transcript provided"""

    assert parser.parse("BOGUS:c.22+1A>T")

    # HGNC approved symbols include non-alphanumeric:
    # dashes     - ADAMTSL4-AS1 or TRL-CAA5-1
    assert parser.parse("BOGUS-EXCELLENT:c.22+1A>T")
    assert parser.parse("BOGUS-MOST-EXCELLENT:c.22+1A>T")

    # underscore - GTF2H2C_2, APOBEC3A_B, C4B_2
    assert parser.parse("BOGUS_EXCELLENT:c.22+1A>T")

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("1BOGUS:c.22+1A>T")  # Starts with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("-BOGUS:c.22+1A>T")  # Starts with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("BOGUS-:c.22+1A>T")  # Ends with non-alpha

    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("BOGUS/EXCELLENT:c.22+1A>T")  # contains invalid character


class Test_Parser(unittest.TestCase):
    longMessage = True

    @classmethod
    def setUpClass(cls):
        cls.parser = hgvs.parser.Parser()

    def test_parser_parse_shorthand(self):
        v = "NM_01234.5:c.22+1A>T"
        assert self.parser.parse_hgvs_variant(v) == self.parser.parse(v)

    def test_parser_gauntlet(self):
        fn = os.path.join(os.path.dirname(__file__), "data", "gauntlet")
        for var in open(fn, "r"):
            var = var.strip()
            if var.startswith("#") or var == "":
                continue
            v = self.parser.parse_hgvs_variant(var)
            self.assertEqual(
                var,
                v.format(conf={"max_ref_length": None}),
                "parse-format roundtrip failed:" + pprint.pformat(v.posedit),
            )

    @pytest.mark.quick
    def test_parser_reject(self):
        fn = os.path.join(os.path.dirname(__file__), "data", "reject")
        for var in open(fn, "r"):
            var, msg = var.strip().split("\t")
            if var.startswith("#") or var == "":
                continue
            with self.assertRaises(HGVSParseError):
                self.parser.parse_hgvs_variant(var)
                self.assertTrue(False, msg="expected HGVSParseError: %s (%s)" % (var, msg))

    @pytest.mark.quick
    def test_parser_posedit_special(self):
        # See note in grammar about parsing p.=, p.?, and p.0
        self.assertEqual(str(self.parser.parse_p_posedit("0")), "0")
        self.assertEqual(str(self.parser.parse_p_posedit("0?")), "0?")
        # self.assertEqual( str(self.parser.parse_p_posedit("(0)")), "0?" )

        self.assertIsNone(self.parser.parse_p_posedit("?"))

        self.assertEqual(str(self.parser.parse_p_posedit("=")), "=")
        # self.assertEqual( str(self.parser.parse_p_posedit("=?")), "(=)" )
        self.assertEqual(str(self.parser.parse_p_posedit("(=)")), "(=)")

    @pytest.mark.quick
    def test_grammar_and_generated_code_in_sync(self):
        """We generate Python code from the OMeta grammar
        This test checks that the grammar file hasn't changed since we generated the Python code"""

        script_path = os.path.realpath(__file__)
        script_dir = os.path.dirname(script_path)
        hgvs_base_dir = os.path.dirname(script_dir)
        grammar_filename = "src/hgvs/_data/hgvs.pymeta"
        generated_filename = "src/hgvs/generated/hgvs_grammar.py"

        # Hash the grammar file
        with open(os.path.join(hgvs_base_dir, grammar_filename), "rb") as grammar_f:
            grammar_hash = hashlib.md5(grammar_f.read()).hexdigest()

        # Read the stored grammar file hash from generated file
        with open(os.path.join(hgvs_base_dir, generated_filename), "r") as generated_f:
            generated_hash = None
            for line in generated_f:
                if not line.startswith("#"):
                    break
                m = re.match(r".*Grammar hash: ([a-fA-F0-9]{32})", line)
                if m:
                    generated_hash = m.group(1)

            msg = "Could not retrieve generated hash from {generated_hash}".format(
                generated_hash=generated_hash
            )
            self.assertIsNotNone(generated_hash, msg)

        msg = (
            "OMeta source '{grammar_filename}' is different than the version used to generate "
            "Python code '{generated_filename}'. You need to run "
            "'sbin/generate_parser.py' ".format(
                grammar_filename=grammar_filename, generated_filename=generated_filename
            )
        )
        self.assertEqual(generated_hash, grammar_hash, msg)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Utility to normalize variants.
Code of the normalization utilities were imported from vgraph
https://github.com/bioinformed/vgraph

"""

from collections import namedtuple

def trim_common_suffixes(strs, min_len=0):
    """
    trim common suffixes

    >>> trim_common_suffixes('A', 1)
    (0, 'A')

    """

    if len(strs) < 2:
        return 0, strs

    rev_strs = [s[::-1] for s in strs]

    trimmed, rev_strs = trim_common_prefixes(rev_strs, min_len)

    if trimmed:
        strs = [s[::-1] for s in rev_strs]

    return trimmed, strs


def trim_common_prefixes(strs, min_len=0):
    """trim common prefixes"""

    trimmed = 0

    if len(strs) > 1:
        s1 = min(strs)
        s2 = max(strs)

        for i in range(len(s1) - min_len):
            if s1[i] != s2[i]:
                break
            trimmed = i + 1

    if trimmed > 0:
        strs = [s[trimmed:] for s in strs]

    return trimmed, strs


def normalize_alleles_left(ref, start, stop, alleles, bound, ref_step, shuffle=True):
    """
    Normalize loci by removing extraneous reference padding

    >>> normalize_alleles_left('A', 1, 2, 'A', 1, 2)
    shuffled_alleles(start=1, stop=2, alleles='A')

    """

    normalized_alleles = namedtuple("shuffled_alleles", "start stop alleles")

    if len(alleles) < 2:
        return normalized_alleles(start, stop, alleles)

    # STEP 1: Trim common suffix
    trimmed, alleles = trim_common_suffixes(alleles)
    stop -= trimmed

    # STEP 2: Trim common prefix
    trimmed, alleles = trim_common_prefixes(alleles)
    start += trimmed

    # assert bound <= start,'start={:d}, left bound={:d}'.format(start, bound)

    # STEP 3: While a null allele exists, left shuffle by prepending alleles
    #         with reference and trimming common suffixes
    while shuffle and "" in alleles and start > bound:
        step = min(ref_step, start - bound)

        r = ref[start - step : start].upper()
        new_alleles = [r + a for a in alleles]

        trimmed, new_alleles = trim_common_suffixes(new_alleles)

        if not trimmed:
            break

        start -= trimmed
        stop -= trimmed

        if trimmed == step:
            alleles = new_alleles
        else:
            left = step - trimmed
            alleles = [a[left:] for a in new_alleles]
            break

    return normalized_alleles(start, stop, tuple(alleles))


def normalize_alleles_right(ref, start, stop, alleles, bound, ref_step, shuffle=True):
    """
    Normalize loci by removing extraneous reference padding

    >>> normalize_alleles_right('A', 1, 2, 'A', 1, 2)
    shuffled_alleles(start=1, stop=2, alleles='A')

    """

    normalized_alleles = namedtuple("shuffled_alleles", "start stop alleles")

    if len(alleles) < 2:
        return normalized_alleles(start, stop, alleles)

    # STEP 1: Trim common prefix
    trimmed, alleles = trim_common_prefixes(alleles)
    start += trimmed

    # STEP 2: Trim common suffix
    trimmed, alleles = trim_common_suffixes(alleles)
    stop -= trimmed

    # assert bound >= stop,'stop={:d}, right bound={:d}'.format(stop, bound)

    # STEP 3: While a null allele exists, right shuffle by appending alleles
    #         with reference and trimming common prefixes
    while shuffle and "" in alleles and stop < bound:
        step = min(ref_step, bound - stop)

        r = ref[stop : stop + step].upper()
        new_alleles = [a + r for a in alleles]

        trimmed, new_alleles = trim_common_prefixes(new_alleles)

        if not trimmed:
            break

        start += trimmed
        stop += trimmed

        if trimmed == step:
            alleles = new_alleles
        else:
            left = step - trimmed
            alleles = [a[:-left] for a in new_alleles]
            break

    return normalized_alleles(start, stop, tuple(alleles))


def normalize_alleles(ref, start, stop, alleles, bound, ref_step, left, shuffle=True):
    if left:
        return normalize_alleles_left(ref, start, stop, alleles, bound, ref_step, shuffle)
    else:
        return normalize_alleles_right(ref, start, stop, alleles, bound, ref_step, shuffle)


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class TxRulerRenderer(BasicRenderer):
    """
    TxRulerRenderer is a class that extends BasicRenderer to provide a display of the transcript position.
    Methods
    -------
    legend() -> str:
        Returns a string representing the legend for the transcript ruler.
    display(data: VariantData) -> str:
        Generates a string that shows the position of the transcript sequence based on the
        provided VariantData. The display method takes into account whether the data is RNA
        and adjusts the position accordingly. It marks positions at intervals of 10 and
        handles cases where positions are not mapped or are at specific intervals.
    """

    def legend(self) -> str:
        return "          : "

    def display(self, data: VariantData) -> str:
        """show the position of the transcript seq"""
        var_str = ""

        is_rna = data.is_rna

        count = -1
        for pdata in data.position_details:
            count += 1
            if not pdata.mapped_pos:
                var_str += " "
                continue

            c_pos = pdata.c_pos
            interval = pdata.c_interval
            if is_rna:
                c_pos = pdata.n_pos
                interval = pdata.n_interval

            if c_pos is None:
                var_str += " "
                continue

            if len(var_str) > count:
                continue

            if (c_pos) % 10 == 0:
                # if pdata.c_interval.start.datum == Datum.CDS_END:
                #     var_str += "*"
                var_str += f"{interval} "
                continue

            elif c_pos == 1:
                var_str += f"{interval} "
                continue
            var_str += " "

        return var_str.rstrip()



# -*- coding: utf-8 -*-
"""Full-featured O(1) LRU cache backported from Python3.3. The full
Py3.3 API is supported (thread safety, maxsize, keyword args, type
checking, __wrapped__, and cache_info). Includes Py3.3 optimizations
for better memory utilization, fewer dependencies, and fewer dict
lookups.

http://code.activestate.com/recipes/578078-py26-and-py30-backport-of-python-33s-lru-cache/

Added persistence capability

"""

from collections import namedtuple
from functools import update_wrapper
from threading import RLock

from ..exceptions import HGVSDataNotAvailableError, HGVSVerifyFailedError

_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])


class _HashedSeq(list):
    __slots__ = ["hashvalue"]

    def __init__(self, tup, hash=hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue

    def __getstate__(self):
        return True  # needs to be truthy for __setstate__ to be called

    def __setstate__(self, _):
        self.hashvalue = hash(tuple(self))

    def __repr__(self):
        return "_HashedSeq({tuple!r})".format(tuple=tuple(self))


def _make_key(
    func,
    args,
    kwds,
    typed,
    kwd_mark=(object(),),
    fasttypes={int, str, frozenset, type(None)},
    sorted=sorted,
    tuple=tuple,
    type=type,
    len=len,
):
    "Make a cache key from optionally typed positional and keyword arguments"
    key = args
    key += kwd_mark
    key += ("__func__", func)
    if kwds:
        sorted_items = sorted(kwds.items())
        for item in sorted_items:
            key += item
    if typed:
        key += tuple(type(v) for v in args)
        if kwds:
            key += tuple(type(v) for k, v in sorted_items)
    elif len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    return _HashedSeq(key)


LEARN = 1
RUN = 2
VERIFY = 3


def lru_cache(maxsize=100, typed=False, mode=None, cache=None):
    """Least-recently-used cache decorator.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    If *typed* is True, arguments of different types will be cached separately.
    For example, f(3.0) and f(3) will be treated as distinct calls with
    distinct results.

    Arguments to the cached function must be hashable.

    View the cache statistics named tuple (hits, misses, maxsize, currsize) with
    f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.

    See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used


    :param mode: cache run mode
        None:   the default lru cache behaver
        LEARN:  queries are executed against persistent cache from file "filename";
                in the event of a miss, the novel query and results would be written to cache, then returned.
        RUN:    queries are executed against caches; misses result in DataNotLocallyAvailableError
        VERIFY: always execute the function; if persistent cache value and returned value are different, raise VerifyFailedError
    :param cache: PersistentDict object or None;

    """  # noqa: E501

    # Users should only access the lru_cache through its public API:
    #       cache_info, cache_clear, and f.__wrapped__
    # The internals of the lru_cache are encapsulated for thread safety and
    # to allow the implementation to change (including a possible C version).

    def decorating_function(user_function):
        lock = RLock()  # because linkedlist updates aren't threadsafe

        _cache = cache
        _maxsize = maxsize

        if _cache is None:
            _cache = dict()
        elif mode is not None:
            _maxsize = None

        stats = [0, 0]  # make statistics updateable non-locally
        HITS, MISSES = 0, 1  # names for the stats fields
        make_key = _make_key
        cache_get = _cache.get  # bound method to lookup key or return None
        _len = len  # localize the global len() function

        root = []  # root of the circular doubly linked list
        root[:] = [root, root, None, None]  # initialize by pointing to self
        nonlocal_root = [root]  # make updateable non-locally
        PREV, NEXT, KEY, RESULT = 0, 1, 2, 3  # names for the link fields

        if _maxsize == 0:

            def wrapper(*args, **kwds):
                # no caching, just do a statistics update after a successful call
                result = user_function(*args, **kwds)
                stats[MISSES] += 1
                return result

        elif _maxsize is None:

            def wrapper(*args, **kwds):
                # simple caching without ordering or size limit
                key = make_key(user_function.__name__, args, kwds, typed, ())
                result = cache_get(key, root)  # root used here as a unique not-found sentinel
                if result is not root:
                    stats[HITS] += 1
                    if mode == VERIFY:
                        latestres = user_function(*args, **kwds)
                        if latestres != result:
                            raise HGVSVerifyFailedError(
                                "The cached result is not consistent with latest result when calling "
                                + user_function.__name__
                                + " with args "
                                + str(args)
                                + " and keywords "
                                + str(kwds)
                            )
                    return result
                if mode == RUN:
                    raise HGVSDataNotAvailableError(
                        "Data not available in local cache when calling "
                        + user_function.__name__
                        + " with args "
                        + str(args)
                        + " and keywords "
                        + str(kwds)
                    )
                result = user_function(*args, **kwds)
                _cache[key] = result
                if mode == LEARN:
                    _cache.sync()
                stats[MISSES] += 1
                return result

        else:

            def wrapper(*args, **kwds):
                # size limited caching that tracks accesses by recency
                key = make_key(user_function.__name__, args, kwds, typed, ())
                with lock:
                    link = cache_get(key)
                    if link is not None:
                        # record recent use of the key by moving it to the front of the list
                        (root,) = nonlocal_root
                        link_prev, link_next, key, result = link
                        link_prev[NEXT] = link_next
                        link_next[PREV] = link_prev
                        last = root[PREV]
                        last[NEXT] = root[PREV] = link
                        link[PREV] = last
                        link[NEXT] = root
                        stats[HITS] += 1
                        return result
                result = user_function(*args, **kwds)
                with lock:
                    (root,) = nonlocal_root
                    if key in _cache:
                        # getting here means that this same key was added to the
                        # cache while the lock was released.  since the link
                        # update is already done, we need only return the
                        # computed result and update the count of misses.
                        pass
                    elif _len(_cache) >= _maxsize:
                        # use the old root to store the new key and result
                        oldroot = root
                        oldroot[KEY] = key
                        oldroot[RESULT] = result
                        # empty the oldest link and make it the new root
                        root = nonlocal_root[0] = oldroot[NEXT]
                        oldkey = root[KEY]
                        root[KEY] = root[RESULT] = None
                        # now update the cache dictionary for the new links
                        del _cache[oldkey]
                        _cache[key] = oldroot
                    else:
                        # put result in a new link at the front of the list
                        last = root[PREV]
                        link = [last, root, key, result]
                        last[NEXT] = root[PREV] = _cache[key] = link
                    stats[MISSES] += 1
                return result

        def cache_info():
            """Report cache statistics"""
            with lock:
                return _CacheInfo(stats[HITS], stats[MISSES], _maxsize, len(_cache))

        def cache_clear():
            """Clear the cache and cache statistics"""
            with lock:
                _cache.clear()
                root = nonlocal_root[0]
                root[:] = [root, root, None, None]
                stats[:] = [0, 0]

        wrapper.__wrapped__ = user_function
        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        return update_wrapper(wrapper, user_function)

    return decorating_function



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper
from hgvs.exceptions import HGVSInvalidVariantError


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign"
        )
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh38", alt_aln_method="splign"
        )
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_260_raise_exception_when_mapping_bogus_variant(self):
        v = self.hp.parse_hgvs_variant("NM_000059.3:c.7790delAAG")
        with self.assertRaises(HGVSInvalidVariantError):
            self.am37.c_to_p(v)

    def test_285_partial_palindrome_inversion(self):
        # https://github.com/biocommons/hgvs/issues/285/
        # Inversion mapping is not palindrome aware
        v = self.hp.parse_hgvs_variant("NM_000088.3:c.589_600inv")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NM_000088.3:c.590_599inv")

    def test_293_parser_attribute_assignment_error(self):
        # https://github.com/biocommons/hgvs/issues/293/
        var = self.hp.parse_hgvs_variant("NG_029146.1:g.6494delG")
        self.vn.normalize(var)  # per issue, should raise error



# -*- coding: utf-8 -*-
"""Utility class that projects variants from one transcript to another
via a common reference sequence.

"""

import copy

import hgvs
import hgvs.alignmentmapper


class Projector:
    """
    The Projector class implements liftover between two transcripts via a
    common reference sequence.

    :param hdp: HGVS Data Provider Interface-compliant instance (see :class:`hgvs.dataproviders.interface.Interface`)
    :param ref: string representing the common reference assembly (e.g., GRCh37.p10)
    :param src_ac: string representing the source transcript accession (e.g., NM_000551.2)
    :param dst_ac: string representing the destination transcript accession (e.g., NM_000551.3)
    :param src_alt_aln_method: string representing the source transcript alignment method
    :param dst_alt_aln_method: string representing the destination transcript alignment method

    This class assumes (and verifies) that the transcripts are on the same
    strand. This assumption obviates some work in flipping sequence
    variants twice unnecessarily.
    """

    def __init__(
        self,
        hdp,
        alt_ac,
        src_ac,
        dst_ac,
        src_alt_aln_method=hgvs.global_config.mapping.alt_aln_method,
        dst_alt_aln_method=hgvs.global_config.mapping.alt_aln_method,
    ):
        self.hdp = hdp
        self.alt_ac = alt_ac
        self.src_tm = hgvs.alignmentmapper.AlignmentMapper(hdp, src_ac, alt_ac, src_alt_aln_method)
        self.dst_tm = hgvs.alignmentmapper.AlignmentMapper(hdp, dst_ac, alt_ac, dst_alt_aln_method)

    def project_interval_forward(self, c_interval):
        """
        project c_interval on the source transcript to the
        destination transcript

        :param c_interval: an :class:`hgvs.interval.Interval` object on the source transcript
        :returns: c_interval: an :class:`hgvs.interval.Interval` object on the destination transcript
        """
        return self.dst_tm.g_to_c(self.src_tm.c_to_g(c_interval))

    def project_interval_backward(self, c_interval):
        """
        project c_interval on the destination transcript to the
        source transcript

        :param c_interval: an :class:`hgvs.interval.Interval` object on the destination transcript
        :returns: c_interval: an :class:`hgvs.interval.Interval` object on the source transcript
        """
        return self.src_tm.g_to_c(self.dst_tm.c_to_g(c_interval))

    def project_variant_forward(self, c_variant):
        """
        project c_variant on the source transcript onto the destination transcript

        :param c_variant: an :class:`hgvs.sequencevariant.SequenceVariant` object on the source transcript
        :returns: c_variant: an :class:`hgvs.sequencevariant.SequenceVariant` object on the destination transcript
        """
        if c_variant.ac != self.src_tm.tx_ac:
            raise RuntimeError(
                "variant accession does not match that used to initialize " + __name__
            )
        new_c_variant = copy.deepcopy(c_variant)
        new_c_variant.ac = self.dst_tm.tx_ac
        new_c_variant.posedit.pos = self.project_interval_forward(c_variant.posedit.pos)
        return new_c_variant

    def project_variant_backward(self, c_variant):
        """
        project c_variant on the source transcript onto the destination transcript

        :param c_variant: an :class:`hgvs.sequencevariant.SequenceVariant` object on the source transcript
        :returns: c_variant: an :class:`hgvs.sequencevariant.SequenceVariant` object on the destination transcript
        """
        if c_variant.ac != self.dst_tm.tx_ac:
            raise RuntimeError(
                "variant accession does not match that used to initialize " + __name__
            )
        new_c_variant = copy.deepcopy(c_variant)
        new_c_variant.ac = self.src_tm.tx_ac
        new_c_variant.posedit.pos = self.project_interval_backward(c_variant.posedit.pos)
        return new_c_variant



# -*- coding: utf-8 -*-
"""Decorator factory class which returns a decorator function that
marks a function as deprecated.

"""

import collections
import warnings


class deprecated:
    """Decorator factory class which returns a decorator function that
    marks a function as deprecated. It will result in a warning being
    emitted when the function is used, once per invocation point and
    deprecated function. The doc string of the deprecated function is
    overwritten with a deprecation message.

    @deprecated(use_instead="c_to_p(...)")
    def hgvsc_to_hgvsp():
        ...

    results in warnings like this:

    variantmapper.py:280: DeprecationWarning: Call to deprecated function
      hgvsc_to_hgvsp; use c_to_p(...) instead

    """

    def __init__(self, use_instead=None):
        # only warn once per caller location and function
        self.seen = collections.Counter()
        self.use_instead = use_instead

    def __call__(self, func):
        msg = "Call to deprecated function {}".format(func.__name__)
        if self.use_instead:
            msg += "; use " + self.use_instead + " instead"

        def wrapper(*args, **kwargs):
            fingerprint = (func.__name__, func.__code__.co_filename, func.__code__.co_firstlineno)
            if fingerprint not in self.seen:
                warnings.warn_explicit(
                    msg,
                    category=DeprecationWarning,
                    filename=func.__code__.co_filename,
                    lineno=func.__code__.co_firstlineno + 1,
                )
            self.seen.update([fingerprint])
            return func(*args, **kwargs)

        wrapper.__doc__ = "Deprecated"
        if self.use_instead:
            wrapper.__doc__ += "; use " + self.use_instead + " instead"
        return wrapper



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs.dataproviders.uta
import hgvs.location
import hgvs.parser
import hgvs.projector


class TestHgvsProjector(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.alt_ac = "NC_000001.10"
        cls.alt_aln_method = "splign"
        cls.hp = hgvs.parser.Parser()

    def tst_forward_and_backward(self, v1, v2):
        pj = hgvs.projector.Projector(
            self.hdp, self.alt_ac, v1.ac, v2.ac, self.alt_aln_method, self.alt_aln_method
        )
        self.assertEqual(pj.project_variant_forward(v1), v2)
        self.assertEqual(pj.project_variant_backward(v2), v1)

    @pytest.mark.quick
    def test_rs201430561(self):
        # rs201430561 http://www.ncbi.nlm.nih.gov/projects/SNP/snp_ref.cgi?rs=201430561
        hgvs_c = ["NM_001197320.1:c.281C>T", "NM_021960.4:c.740C>T", "NM_182763.2:c.688+403C>T"]
        var_c = [self.hp.parse_hgvs_variant(h) for h in hgvs_c]
        self.tst_forward_and_backward(var_c[0], var_c[1])
        self.tst_forward_and_backward(var_c[0], var_c[2])
        self.tst_forward_and_backward(var_c[1], var_c[2])

    @pytest.mark.quick
    def test_bad_acs(self):
        hgvs_c = ["NM_001197320.1:c.281C>T", "NM_021960.4:c.740C>T", "NM_182763.2:c.688+403C>T"]
        var_c = [self.hp.parse_hgvs_variant(h) for h in hgvs_c]
        pj = hgvs.projector.Projector(
            self.hdp,
            self.alt_ac,
            var_c[0].ac,
            var_c[1].ac,
            self.alt_aln_method,
            self.alt_aln_method,
        )
        # intentionally call p_v_f with variant on *destination* transcript, and vice versa
        with self.assertRaises(RuntimeError):
            pj.project_variant_forward(var_c[1])
        with self.assertRaises(RuntimeError):
            pj.project_variant_backward(var_c[0])


if __name__ == "__main__":
    unittest.main()



import re

from hgvs.enums import ValidationLevel

g_ac = r"|".join([r"^(?:NC_|NG_|NT_|NW_|NZ_|CM)\d+\.\d+$", r"^LRG_\d+$"])
cr_ac = r"|".join([r"^(?:NM_|XM_|ENST)\d+\.\d+$", r"^LRG_\d+t\d+$"])
n_ac = r"|".join([r"^(?:NR_|XR_|ENST)\d+\.\d+$", r"^LRG_\d+t\d+$"])
p_ac = r"|".join([r"^(?:NP_|XP_|ENSP)\d+\.\d+$", r"^LRG_\d+p\d+$"])
m_ac = r"|".join([r"^(?:NC_)\d+\.\d+$"])

valid_pairs = {
    "c": re.compile(cr_ac),
    "g": re.compile(g_ac),
    "m": re.compile(m_ac),
    "n": re.compile(n_ac),
    "p": re.compile(p_ac),
    "r": re.compile(cr_ac),
}

invalid_pairs = {
    "c": re.compile(r"|".join((g_ac, p_ac))),
    "g": re.compile(r"|".join((cr_ac, n_ac, p_ac))),
    "m": re.compile(r"|".join((cr_ac, n_ac, p_ac))),
    "n": re.compile(r"|".join((g_ac, p_ac))),
    "p": re.compile(r"|".join((g_ac, cr_ac, n_ac))),
    "r": re.compile(r"|".join((g_ac, p_ac))),
}


def validate_type_ac_pair(type, ac):
    """validate that accession is correct for variant type AND that
    accession is fully specified.

    """

    assert type in valid_pairs, "Unknown variant type " + type
    if valid_pairs[type].match(ac):
        return (
            ValidationLevel.VALID,
            "Accession ({ac}) is compatible with variant type {type}".format(ac=ac, type=type),
        )
    elif invalid_pairs[type].match(ac):
        return (
            ValidationLevel.ERROR,
            "Accession ({ac}) is not compatible with variant type {type}".format(ac=ac, type=type),
        )
    else:
        return (
            ValidationLevel.WARNING,
            "Accession ({ac}) is not known to be compatible with variant type {type}".format(
                ac=ac, type=type
            ),
        )



from bioutils.normalize import normalize


def test_shuffling():
    ref = "TAAAAAAAT"
    alt = "TATAAAAAAT"

    chrom_seq = ref
    start = 0
    end = len(ref)
    shuffle_direction = "EXPAND"

    shuffled_interval, shuffled_alleles = normalize(
        chrom_seq, interval=(start, end), alleles=(None, alt), mode=shuffle_direction
    )

    assert shuffled_interval == (2, 2)
    assert shuffled_alleles[0] == ""
    assert shuffled_alleles[1] == "T"



import os
from contextlib import contextmanager

import hgvs
from hgvs.exceptions import HGVSError
import pytest
import support.mock_input_source as mock_input_data_source

sanity_cases = [
    {
        "name": "ins at intron/exon boundary",
        "var_c": "NM_999999.1:c.2-1_2insG",
        "exonic":{
            "var_p": "MOCK:p.Met1?",
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },
    },
    {
        "name": "ins at exon/intron boundary",
        "var_c": "NM_999999.1:c.26_26+1insC",
        "exonic":{
            "var_p": "MOCK:p.(Lys9AsnfsTer?)",
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary",
        "var_c": "NM_999999.1:c.2dup",
        "exonic":{
            "var_p": "MOCK:p.Met1?",
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary",
        "var_c": "NM_999999.1:c.2-1dup",
        "exonic":{
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },

    },
    {
        "name": "dup at exon/intron boundary",
        "var_c": "NM_999999.1:c.26dup",
        "exonic":{
            "var_p": "MOCK:p.(Ter10IleextTer?)",
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary",
        "var_c": "NM_999999.1:c.26+1dup",
        "exonic":{
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "MOCK:p.?",
        },
    },
]

real_cases = [
    {
        "name": "ins at intron/exon boundary",
        "var_c": "NM_004380.2:c.3251-1_3251insA",
        "exonic":{
            "var_p": "NP_004371.2:p.(Ile1084AsnfsTer3)",
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "ins at exon/intron boundary",
        "var_c": "NM_004380.2:c.3250_3250+1insT",
        "exonic":{
            "var_p": "NP_004371.2:p.(Phe1085LeufsTer2)",
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in exon, single-base)",
        "var_c": "NM_024529.4:c.132dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45AspfsTer21)",
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in exon, multi-base)",
        "var_c": "NM_024529.4:c.132_133dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45ArgfsTer65)",
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in intron, single-base)",
        "var_c": "NM_024529.4:c.132-1dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45AspfsTer21)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in intron, multi-base)",
        "var_c": "NM_024529.4:c.132-2_132-1dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45GlyfsTer65)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in exon, single-base, negative strand)",
        "var_c": "NM_004380.2:c.3251dup",
        "exonic":{
            "var_p": "NP_004371.2:p.(Phe1085LeufsTer2)",
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in exon, multi-base, negative strand)",
        "var_c": "NM_004380.2:c.3251_3252dup",
        "exonic":{
            "var_p": "NP_004371.2:p.(Phe1085SerfsTer15)",
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in intron, single-base, negative strand)",
        "var_c": "NM_004380.2:c.3251-1dup",
        "exonic":{
            "var_p": "NP_004371.2:p.(Ile1084SerfsTer3)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "dup at intron/exon boundary (in intron, multi-base, negative strand)",
        "var_c": "NM_004380.2:c.3251-2_3251-1dup",
        "exonic":{
            "var_p": "NP_004371.2:p.(Ile1084LysfsTer16)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_004371.2:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in exon, single-base)",
        "var_c": "NM_024529.4:c.131dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45AspfsTer21)",
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in exon, multi-base)",
        "var_c": "NM_024529.4:c.130_131dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45GlyfsTer65)",
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in intron, single-base)",
        "var_c": "NM_024529.4:c.131+1dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45AspfsTer21)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in intron, multi-base)",
        "var_c": "NM_024529.4:c.131+1_131+3dup",
        "exonic":{
            "var_p": "NP_078805.3:p.(Thr45Ter)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_078805.3:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in exon, single-base, negative strand)",
        "var_c": "NM_004985.4:c.111dup",
        "exonic":{
            "var_p": "NP_004976.2:p.(Asp38GlyfsTer10)",
        },
        "intronic":{
            "var_p": "NP_004976.2:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in exon, multi-base, negative strand)",
        "var_c": "NM_004985.4:c.110_111dup",
        "exonic":{
            "var_p": "NP_004976.2:p.(Asp38ArgfsTer8)",
        },
        "intronic":{
            "var_p": "NP_004976.2:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in intron, single-base, negative strand)",
        "var_c": "NM_004985.4:c.111+1dup",
        "exonic":{
            "var_p": "NP_004976.2:p.(Asp38GlyfsTer10)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_004976.2:p.?",
        },
    },
    {
        "name": "dup at exon/intron boundary (in intron, multi-base, negative strand)",
        "var_c": "NM_004985.4:c.111+1_111+4dup",
        "exonic":{
            "var_p": "NP_004976.2:p.(Asp38ValfsTer11)",
            "exc": HGVSError,
        },
        "intronic":{
            "var_p": "NP_004976.2:p.?",
        },
    },
    {
        "name": "snv with broken cigar mapping",
        "var_c": "NM_014638.2:c.515-99G>T",
        "exonic":{
            "var_p": "NP_055453.2:p.?",
        },
        "intronic":{
            "var_p": "NP_055453.2:p.?",
        },
    },
    {
        "name": "ins with broken cigar mapping",
        "var_c": "NM_004799.2:c.71-7048_71-7047insATAT",
        "exonic":{
            "var_p": "NP_004790.2:p.?",
        },
        "intronic":{
            "var_p": "NP_004790.2:p.?",
        },
    },
    {
        "name": "del with broken cigar mapping",
        "var_c": "NM_007324.2:c.71-7826_71-7802del",
        "exonic":{
            "var_p": "NP_015563.2:p.?",
        },
        "intronic":{
            "var_p": "NP_015563.2:p.?",
        },
    },
]


@pytest.fixture(scope="module")
def mock_hdp():
    fn = os.path.join(os.path.dirname(__file__), "..", "data", "sanity_cp.tsv")
    return mock_input_data_source.MockInputSource(fn)


@pytest.fixture(scope="module")
def mock_vm(mock_hdp):
    return hgvs.variantmapper.VariantMapper(mock_hdp, prevalidation_level="INTRINSIC")


@contextmanager
def config_setting(module, setting, temp_value):
    old_value = getattr(module, setting)
    try:
        setattr(module, setting, temp_value)
        yield
    finally:
        setattr(module, setting, old_value)


@pytest.mark.parametrize("case", sanity_cases)
def test_sanity_c_to_p(case, parser, mock_vm):
    with config_setting(hgvs.global_config.mapping, 'ins_at_boundary_is_intronic', False):
        var_c = parser.parse(case["var_c"])
        if "exc" in case["exonic"]:
            with pytest.raises(case["exonic"]["exc"]):
                mock_vm.c_to_p(var_c, "MOCK")
        else:
            assert str(mock_vm.c_to_p(var_c, "MOCK")) == case["exonic"]["var_p"]
    with config_setting(hgvs.global_config.mapping, 'ins_at_boundary_is_intronic', True):
        var_c = parser.parse(case["var_c"])
        if "exc" in case["intronic"]:
            with pytest.raises(case["intronic"]["exc"]):
                mock_vm.c_to_p(var_c, "MOCK")
        else:
            assert str(mock_vm.c_to_p(var_c, "MOCK")) == case["intronic"]["var_p"]


@pytest.mark.parametrize("case", real_cases)
def test_real_c_to_p(case, parser, vm, am37):
    with config_setting(hgvs.global_config.mapping, 'ins_at_boundary_is_intronic', False):
        var_c = parser.parse(case["var_c"])
        if "exc" in case["exonic"]:
            with pytest.raises(case["exonic"]["exc"]):
                vm.c_to_p(var_c)
        else:
            assert str(vm.c_to_p(var_c)) == case["exonic"]["var_p"]
        assert str(am37.c_to_p(var_c)) == case["exonic"]["var_p"]
    with config_setting(hgvs.global_config.mapping, 'ins_at_boundary_is_intronic', True):
        var_c = parser.parse(case["var_c"])
        if "exc" in case["intronic"]:
            with pytest.raises(case["intronic"]["exc"]):
                vm.c_to_p(var_c)
        else:
            assert str(vm.c_to_p(var_c)) == case["intronic"]["var_p"]
        assert str(am37.c_to_p(var_c)) == case["intronic"]["var_p"]
